#ifndef VEKTOR_H
#define VEKTOR_H
#include <algorithm>                        // copy() usw.
#include <initializer_list>
#include <memory>                           // unique_ptr
#include <stdexcept>
#include <utility>                          // swap()

template <typename T> 
class [[nodiscard]] Vektor {
public:
  explicit Vektor(int anzahl = 1);         // Allg. Konstruktor
  Vektor(int n, T t);               // 2. Allg. Konstruktor: n Elemente mit Wert t
  Vektor(std::initializer_list<T>);         // Sequenzkonstruktor
  Vektor(const Vektor& v);                  // Kopierkonstruktor

  [[nodiscard]] auto size() const { return anzahl; }

  void groesseAendern(int);                 // dynamisch ändern

  void checkIndex(int index) const          // Index prüfen
  {
    if (index >= anzahl) {
      throw std::out_of_range("Indexüberschreitung!");
    }
  }

  T& operator[](int index)                  // Indexoperator
  {
    checkIndex(index);
    return start[index];
  }

  const T& operator[](int index) const       // für nicht-veränderliche Vektoren
  {
    checkIndex(index);
    return start[index];
  }

  // Zuweisungsoperator
  Vektor& operator=(Vektor);             // Übergabe per Wert, siehe Impl. unten
  void swap(Vektor& v) noexcept;                  // Vektoren vertauschen

  // Zeiger auf Anfang und Position nach dem Ende für Vektoren
  // mit nichtkonstanten und konstanten Elementen
  [[nodiscard]] T* begin()             { return start.get(); }
  [[nodiscard]] T* end()               { return start.get() + anzahl; }
  [[nodiscard]] const T* begin() const { return start.get(); }
  [[nodiscard]] const T* end() const   { return start.get() + anzahl; }

private:
  int anzahl;                                // Anzahl der Datenobjekte
  std::unique_ptr<T[]> start;                // Zeiger auf Datenobjekte
};

/********** Implementierung ****************/

template <typename T>
Vektor<T>::Vektor(int anz)          // erster Allg. Konstruktor
  : anzahl{anz}                     // unique_ptr ist leer
{
  if(anz < 1) {
    throw std::logic_error("Zeilen-/Spaltenzahl muss > 0 sein!");
  }
  start.reset(new T[anz]);
}

template <typename T>
Vektor<T>::Vektor(int anz, T wert)          // Allg. Konstruktor 2
    : Vektor(anz)           // Delegation an obigen Konstruktor
{
  std::fill(begin(), begin() + anzahl, wert);   // statt Schleife
}

template <typename T>
Vektor<T>::Vektor(std::initializer_list<T> liste)          // Sequenzkonstruktor
    : Vektor(liste.size())           // Delegation an obigen Konstruktor
{
  std::copy(liste.begin(), liste.end(), begin());
}

template <typename T>
Vektor<T>::Vektor(const Vektor& v)          // Kopierkonstruktor
    : Vektor(v.anzahl)
{
  std::copy_n(v.begin(), anzahl, begin());  // statt Schleife
}

template <typename T> 
void Vektor<T>::swap(Vektor& v) noexcept
{
  std::swap(anzahl, v.anzahl);
  std::swap(start, v.start);
}

// Alternative mit 'swap-Trick'
template <typename T> 
Vektor<T>& Vektor<T>::operator=(Vektor v)        // Zuweisung
{
  swap(v);
  return *this;
}

template <typename T>
void Vektor<T>::groesseAendern(int neueGroesse)
{
  auto pTemp {std::make_unique<T[]>(neueGroesse)};   // neuen Speicherplatz besorgen
  // die richtige Anzahl von Elementen kopieren
  int kleinereZahl{neueGroesse > anzahl ? anzahl : neueGroesse};
  std::copy_n(begin(), kleinereZahl, pTemp.get());  // statt Schleife
  start.swap(pTemp);                 // Verwaltungsdaten aktualisieren
  anzahl = neueGroesse;              // Verwaltungsdaten aktualisieren
}
#endif

